/* This file is part of SLB
   Copyright (C) 2011 Sergey Poznyakoff

   SLB is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SLB is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SLB.  If not, see <http://www.gnu.org/licenses/>. */

#include "slb.h"

static Hash_table *expr_table;

static size_t
expr_hasher (void const *data, unsigned n_buckets)
{
	struct slb_expression const *exp = data;
	return hash_string(exp->ex_name, n_buckets);
}

/* Compare two strings for equality.  */
static bool
expr_compare(void const *data1, void const *data2)
{
	struct slb_expression const *exp1 = data1;
	struct slb_expression const *exp2 = data2;
	return strcmp(exp1->ex_name, exp2->ex_name) == 0;
}

static void
expr_free(void *data)
{
	struct slb_expression *exp = data;
	/* FIXME: slb_tree_free(exp->ex_tree);*/
	/* FIXME: Free string values */
	free(exp);
}


/* Register an expression. */
int
register_expression(struct slb_expression *expr)
{
	struct slb_expression *s;

	if (!((expr_table
	       || (expr_table = hash_initialize(0, 0,
						expr_hasher,
						expr_compare,
						expr_free)))
	      && (s = hash_insert(expr_table, expr))))
		xalloc_die();

	if (s != expr) {
		grecs_error (&expr->ex_locus, 0,
			     _("redefinition of expression `%s'"),
			     expr->ex_name);
		grecs_error (&s->ex_locus, 0,
			     _("previously defined here"));
		expr_free(expr);
		return 1;
	}
	return 0;
}

struct slb_expression *
find_expression(const char *name)
{
	struct slb_expression ex;
	
	if (!expr_table)
		return NULL;
	ex.ex_name = name;
	return hash_lookup(expr_table, &ex);
}

void
copy_expression(struct slb_expression *src, struct slb_expression **dst)
{
	*dst = xmalloc(sizeof(**dst));
	memcpy(*dst, src, sizeof(**dst));
	(*dst)->state = NULL;
}


static void
nodefixup(struct slb_node *node, struct slb_varinst_tab *vit,
	  struct slb_expression *expr)
{
	struct slb_varinstance *vi;
	struct slb_node *np;
	
	switch (node->n_type) {
	case node_unknown:
		abort();
		
	case node_variable:
		vi = varinst_lookupz(vit,
				     expr->ex_vartab.vt_tab[node->v.n_var].ve_name, 0);
		if (vi) {
			expr->ex_vartab.vt_tab[node->v.n_var].ve_inst = vi;
			vi->vi_flags |= SLV_VIF_USED;
		}
		break;

	case node_uminus:
	case node_not:
		nodefixup(node->v.n_arg[0], vit, expr);
		break;
		
	case node_add:
	case node_sub:
	case node_mul:
	case node_div:
	case node_or:
	case node_and:
	case node_eq:
	case node_ne:
 	case node_lt:
	case node_le:
	case node_gt:
	case node_ge:
		nodefixup(node->v.n_arg[0], vit, expr);
		nodefixup(node->v.n_arg[1], vit, expr);
		break;
		
	case node_cond:
		nodefixup(node->v.n_arg[0], vit, expr);
		if (node->v.n_arg[1])
			nodefixup(node->v.n_arg[1], vit, expr);
		nodefixup(node->v.n_arg[2], vit, expr);
		break;

	case node_func:
		for (np = node->v.n_func.args; np; np = np->n_next)
			nodefixup(np, vit, expr);
		break;

	case node_const:
		break;
	}
}
		
int
fixup_expression(struct slb_expression *expr, struct slb_varinst_tab *vit)
{
	int i;
	int rc = 0;
	
	nodefixup(expr->ex_tree, vit, expr);

	for (i = 0; expr->ex_vartab.vt_max &&
		     expr->ex_vartab.vt_tab[i].ve_name; i++) {
		if (!expr->ex_vartab.vt_tab[i].ve_inst) {
			grecs_error(&expr->ex_locus, 0,
				    _("variable or constant %s not defined"),
				    expr->ex_vartab.vt_tab[i].ve_name);
			rc = 1;
		}
	}

	for (i = 0; vit->vit_max && vit->vit_tab[i].vi_name; i++) {
		if (!(vit->vit_tab[i].vi_flags & SLV_VIF_USED)) {
			if (is_output_refname(vit->vit_tab[i].vi_name))
				vit->vit_tab[i].vi_flags |= SLV_VIF_USED;
			else
				grecs_warning(&expr->ex_locus, 0,
					      _("variable %s is never used"),
					      vit->vit_tab[i].vi_name);
		}
	}
	return rc;
}
